unit tw16ex4u;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, Spin, DB, DBTables, Calendar, Grids, DBGrids,
  Buttons, Menus, ExtCtrls,
  Slider, waveIn, waveOut, IOfile;

type
  TMAIN = class(TForm)
    cbMonth: TComboBox;
    UpDown1: TSpinButton;
    eYear: TEdit;
    DBGrid1: TDBGrid;
    clndr: TCalendar;
    Table1: TTable;
    DataSource1: TDataSource;
    Table1Date: TDateField;
    OpenDialog1: TOpenDialog;
    btSetFormat: TButton;
    BitBtn1: TBitBtn;
    btHelp: TBitBtn;
    btClear: TBitBtn;
    btRecord: TSpeedButton;
    btPause: TSpeedButton;
    btPlay: TSpeedButton;
    btDel: TSpeedButton;
    log: TMemo;
    cbActivation: TCheckBox;
    Panel1: TPanel;
    paInfo: TPanel;
    paStatus: TPanel;
    paFormat: TPanel;
    Label1: TLabel;
    Table1Format: TBytesField;
    Table1Audio: TBlobField;
    Table1ACMFormat: TStringField;
    MainMenu1: TMainMenu;
    muAudio: TMenuItem;
    muSetFormat: TMenuItem;
    N1: TMenuItem;
    muRecord: TMenuItem;
    muPause: TMenuItem;
    muPlay: TMenuItem;
    N2: TMenuItem;
    muExit: TMenuItem;
    muDB: TMenuItem;
    muDel: TMenuItem;
    muLoadFromFile: TMenuItem;
    Label2: TLabel;
    Label3: TLabel;
    btLoad: TBitBtn;
  {startup}
    procedure FormCreate(Sender: TObject);
    procedure initDemoComponents(Sender: TObject);
    procedure initCalender(Sender: TObject);
    procedure setACMFormatString;

 {calender and calender control:}
    procedure cbMonthChange(Sender: TObject);
    procedure eYearChange(Sender: TObject);
    procedure dbGrid1DblClick(Sender: TObject);
    procedure clndrChange(Sender: TObject);
    procedure UpDown1DownClick(Sender: TObject);
    procedure UpDown1UpClick(Sender: TObject);

 {UI - buttons and checkbox}
    procedure btSetFormatClick(Sender: TObject);
    procedure btRecordClick(Sender: TObject);
    procedure btPauseClick(Sender: TObject);
    procedure btPlayClick(Sender: TObject);
    procedure btDelClick(Sender: TObject);
    procedure btClearClick(Sender: TObject);
    procedure btHelpClick(Sender: TObject);
    procedure StopPlay(stMsg : string);
    procedure cbActivationClick(Sender: TObject);
 {UI - sliders}
    procedure tbSpeakerVolChange(Sender: TObject);
    procedure tbActivationLevelChange(Sender: TObject);
 {UI - menu}
    procedure muSetFormatClick(Sender: TObject);
    procedure muRecordClick(Sender: TObject);
    procedure muPauseClick(Sender: TObject);
    procedure muPlayClick(Sender: TObject);
    procedure muDelClick(Sender: TObject);
    procedure muLoadFromFileClick(Sender: TObject);

 {these are the core procedures to study for this demo:}
    procedure makeRecord(table: TTable);       {open a DB record, create stream}
    procedure getAudioFormat(table: TTable);   {read wave format from DB record}
    procedure getAudioData(table: TTable);     {read blob data from DB record}
    procedure LoadfromfileClick(Sender: TObject); {load from wave file to DB record}

 {and the twave component events:}
    procedure WaveInput1Mute(Sender: TObject);
    procedure WaveInput1Overflow(Sender: TObject);
    procedure WaveInput1WaveInput(Sender: TObject);
    procedure WaveOutput1WaveOutput(Sender: TObject);
    procedure WaveOutput1End(Sender: TObject);
 {also, look at btPlayClick, and btRecord click}

  private
    { Private declarations }
  public
    { Public declarations }
    {these components are created at runtime,
    so that you don't have to install them to
    the VCL pallet in order to run the demo}
    waveInput1 : TWaveInput;
    waveOutput1 : TWaveOutput;
    tbSpeakerVol : TSlider;
    tbActivationLevel : TSlider;
  end;

const ID : longint = 1;
var
  MAIN: TMAIN;
  BStream : TBlobStream;  {stream to/from twave-ringbuffer->blob field}
  buffer : pointer;       {waveEvent blocks to/from stream}
  buffersize : longint;   {size of waveEvents for current format}
  cursize : longint;      {size of current blob}
implementation

{$R *.DFM}
procedure TMAIN.FormCreate(Sender: TObject);
begin
  {create demo components at runtime}
   initDemoComponents(self);
   initCalender(self);
   setAcmFormatString;
  {query the audio device for the current volume setting,
    set the slider control to this value}
   WaveOutput1.Open;
   tbSpeakerVol.value:=integer(WaveOutput1.GetVolume) div 655;
   WaveOutput1.Close;
end;

procedure TMAIN.initDemoComponents(Sender: TObject);
begin
(* waveInput1 and waveOutput1 create and initialization code
would not be required if the components were installed to the
Delphi VCL pallet.  The components are created here, in the
form create method, to make testing of these demo projects possible
without having first to install the components.*)
waveInput1 := TwaveInput.create(self);
with waveInput1 do begin
    DeviceID := -1;
    Capacity := 1000;
    Asynchronous := False;
    SamplesPerSec := 11025;
    BytesPerSec := 22050;
    Channels := 1;
    BlockAlign := 2;
    BitsPerSample := 16;
    AutoActivation := True;
    ActivationLevel := 3;
    preFrame := False;
    postFrame := 0;
    BlockSize := 12500;
    OnMute := WaveInput1Mute;
    OnOverflow := WaveInput1Overflow;
    OnWaveInput := WaveInput1WaveInput;
    Left := 80;
    Top := 72;
end;
waveOutput1 := TwaveOutput.create(self);
with waveOutput1 do begin
    DeviceID := -1;
    Capacity := 1000;
    SamplesPerSec := 11025;
    BytesPerSec := 22050;
    Channels := 1;
    BitsPerSample := 16;
    OnEnd := WaveOutput1End;
    OnWaveOutput := WaveOutput1WaveOutput;
    Left := 80;
    Top := 72;
end;
{Slider is not part of Twave; it is a unauthorized hack modification
of slider.pas by Colin Wilson - (woozle@cix.compulink.co.uk);
It serves for this demo, although the paint method does not work
properly under any but the default orientation and size property settings.
It is for Delphi 1.0 Twave demo purposes only.}
tbActivationLevel := TSlider.create(self);
with tbActivationLevel do begin
     parent := self;
     top :=  264;
     left := 96;
     OnValueChange := tbActivationLevelChange;
end;

tbSpeakerVol := TSlider.create(self);
with tbSpeakerVol do begin
     parent := self;
     top :=  296;
     left := 96;
     selHighColor := clRed;
     OnValueChange := tbSpeakerVolChange;
end;
end;

procedure TMAIN.setACMFormatString;
var stFTag   : string[10];
    stChannels : string[5];
begin
  If paFormat.caption = '' then
     begin
     with WaveOutput1 do begin
       if FWF.wFormatTag = 1 then
          stFTag := 'PCM ' else
          stFTag := '?';

       case channels of
            1 : stChannels := 'Mono';
            2 : stChannels := 'Stereo';
            4 : stChannels := 'Quad';
            else stChannels := 'Error';
       end;

        paFormat.caption := stFTag+IntToStr(SamplesPerSec)+' Hz, '+
                         IntToStr(BitsPerSample)+' bits, '+stChannels;
       end;
     end;
end;

procedure TMAIN.initCalender(Sender: TObject);
var i : integer;
begin
   cbMonth.Items.Clear;
   for i:=1 to 12 do cbMonth.items.Add(LongMonthNames[i]);
   cbMonth.ItemIndex:=clndr.Month-1;
   eYear.text:=IntToStr(clndr.Year);
   ClndrChange(Sender);
end;

procedure TMAIN.cbMonthChange(Sender: TObject);
begin
   clndr.Month:=cbMonth.ItemIndex+1;
end;

procedure TMAIN.eYearChange(Sender: TObject);
begin
   clndr.Year:=StrToInt(eYear.text);
end;

procedure TMAIN.dbGrid1DblClick(Sender: TObject);
begin
  if Sender=Clndr then
  if not Table1.FindKey([Clndr.CalendarDate]) then exit;
  btPlay.Down:=true;
  btPlayClick(Sender);
end;

procedure TMAIN.clndrChange(Sender: TObject);
begin
if Table1.FindKey([Clndr.CalendarDate]) then
  btPlay.Enabled:=true
else
  btPlay.Enabled:=false
end;

procedure TMAIN.UpDown1DownClick(Sender: TObject);
begin
  eYear.text := IntToStr(StrToInt(eYear.text)-1)
end;

procedure TMAIN.UpDown1UpClick(Sender: TObject);
begin
  eYear.text := IntToStr(StrToInt(eYear.text)+1)
end;

procedure TMAIN.btSetFormatClick(Sender: TObject);
var wiError, woError : word;
    stWaveFormat : string;

begin
 (*show the ACM format dialog, showing the wave formats installed
   on the current system; note, some formats shown may not be support
   by the installed audio device; see below for error checking*)
   stWaveFormat := WaveInput1.SelectFormat;
   if (stWaveFormat = 'Cancel') or (stWaveFormat = '') then exit;

   {display the selected format}
   paFormat.caption := stWaveFormat;

   {not all formats displayed by ACM are supported, so we still need to test
    if we have a valid format for the system:}
      wiError:=WaveInput1.IsFormatSupported(WaveInput1.PFWF);
      woError:=WaveOutput1.IsFormatSupported(WaveOutput1.PFWF);
   {a returned value of 0 means that the format is supported}

      if ((wiError <> 0) or (woError <> 0)) then
         begin
           paInfo.caption := 'format not supported';
           btRecord.enabled := false;
           btPlay.enabled := false;
         end else
         begin
           waveInput1.clearRingBuffer;
           paInfo.caption := 'format supported';
           btRecord.enabled := true;
           btPlay.enabled := true;
         end;
end;

procedure TMAIN.btRecordClick(Sender: TObject);
begin
   if btPause.down then btPause.down := false;

   if btRecord.Down then begin
      makeRecord(table1); {record opened, stream created}
      {if the installed wave device is full duplex, then it would not
      be *required* to close output begining input}
      btPlay.Down:=false;
      WaveOutput1.Stop;
      WaveOutput1.Close;

      paStatus.caption := 'Recording';
      WaveInput1.Open;
      WaveInput1.Start;

   end else begin  {Record button is going from down to up position}
      paStatus.caption := 'Recording stopped';
      tbActivationLevel.SelHighEnd:=0;    {reset the voice-level indication}
      WaveInput1.Stop;                {stop recording}
      WaveInput1.Close;
      BStream.Free;                   {release the stream created in MakeRecord}
      Table1.Post;                    {post the record}
      btPlay.enabled := true;
   end;
end;

procedure TMAIN.btPauseClick(Sender: TObject);
begin
  if btPause.down then
    begin
       if btRecord.down then
          begin
          paStatus.caption := 'Recording paused';
          tbActivationLevel.SelHighEnd:=0;
          waveInput1.stop;
          waveInput1.close;
          end else
          if btPlay.down then
             begin
             paStatus.caption := 'Playing paused';
             waveOutput1.pause
             end else
               btPause.down := false;
    end else
    begin
       if btRecord.down then
          begin
          paStatus.caption := 'Recording resumed';
          waveInput1.open;
          waveInput1.start;
          end else
          if btPlay.down then
             begin
             paStatus.caption := 'Playing Resumed';
             waveOutput1.resume
             end else
               btPause.down := false;
    end;
end;

procedure TMAIN.btPlayClick(Sender: TObject);
begin
   if btPause.down then btPause.down := false;
   If table1.recordCount = 0 then begin
      btPlay.down := false;
      exit;
   end;
   if not btPlay.Down then
      StopPlay('play stopped') else  {btPlay toggled from down to up}
   begin
      paStatus.caption := 'Playing';
      if waveInput1.state = wisOpen then begin
         btRecord.down:=false;
         btRecordClick(Sender); {run record-stop code}
      end;
      getAudioFormat(table1); {determine format of stored audio,}
                              {and set waveOutput to support conversion}
      getAudioData(table1);  {read the data (if any) from the current db record}
      if WaveOutput1.MessageCount > 0 then
         begin
           Clndr.Enabled:=false;
           DBGrid1.Enabled:=false;
           WaveOutput1.PlayBuffer;
           WaveOutput1.PlayBuffer;
           {playing the buffer will result in a WaveOutput event, where
            you playbuffer will be called again, and thus repeated until the
            waveOutput buffer is empty}
         end else
            StopPlay('no audio data'); {btPlay toggled down, but no content to play}
  end;
end;

procedure TMAIN.btDelClick(Sender: TObject);
begin
  if (waveInput1.state = wisOpen) or
     (waveOutput1.state = wosOpen)or
     (btPause.down) then
     begin
      MessageDlg('End record or play before deleting a record', mtInformation,
      [mbOk], 0) ;
      exit;
     end;
  if Table1.recordcount = 0 then
  begin
      MessageDlg('End record or play before deleting a record', mtInformation,
      [mbOk], 0);
      exit;
  end;
  Table1.Delete;
  paStatus.caption := 'Record deleted';

end;

procedure TMAIN.btClearClick(Sender: TObject);
begin
log.clear;
end;

procedure TMAIN.btHelpClick(Sender: TObject);
begin
application.helpJump('');
end;

procedure TMAIN.StopPlay(stMsg : string);
begin
  {stop playing}
  WaveOutput1.Stop;
  WaveOutput1.Close;
  {clean up}
  Waveoutput1.ClearRingBuffer;
  BStream.Free;
  FreeMem(buffer,buffersize);
  {refresh the UI}
  paStatus.caption := stMsg;
  btPlay.down := false;
  btPause.down := false;
  Clndr.Enabled:=true;
  DBGrid1.Enabled:=true;
end;

procedure TMAIN.cbActivationClick(Sender: TObject);
begin
   waveInput1.autoActivation := cbActivation.checked;
end;

procedure TMAIN.tbSpeakerVolChange(Sender: TObject);
var W : Word;
begin
    W:=tbSpeakerVol.value*655;
    WaveOutput1.SetVolume(W,W);
    tbSpeakerVol.selHighEnd := tbSpeakerVol.value + 5;
end;

procedure TMAIN.tbActivationLevelChange(Sender: TObject);
begin
WaveInput1.ActivationLevel:=tbActivationLevel.value;
end;



procedure TMAIN.muSetFormatClick(Sender: TObject);
begin
  btSetFormatClick(self);
end;

procedure TMAIN.muRecordClick(Sender: TObject);
begin
  btRecord.down := not btRecord.down;
  btRecordClick(Self);
end;

procedure TMAIN.muPauseClick(Sender: TObject);
begin
  btPause.down := not btPause.down;
  btPauseClick(Self);
end;

procedure TMAIN.muPlayClick(Sender: TObject);
begin
  btPlay.down := not btPlay.down;
  btPlayClick(Sender);
end;

procedure TMAIN.muDelClick(Sender: TObject);
begin
  btDelClick(Self);
end;

procedure TMAIN.muLoadFromFileClick(Sender: TObject);
begin
  LoadfromfileClick(self);
end;

procedure TMAIN.LoadfromfileClick(Sender: TObject);
var size : longint;
    buf : array[0..1023] of byte;
begin
  if not OpenDialog1.Execute then exit;
  if Table1.FindKey([Clndr.CalendarDate]) then Table1.Delete;
  size:=OpenWaveFile( OpenDialog1.FileName,
                      WaveOutput1.PFWF,
                      WaveOutput1.sizeFWF);

  Table1.Append;
  Table1.FieldByName('Date').asString:=DateToStr(Clndr.CalendarDate);
  Table1.FieldByName('Format').SetData(WaveOutput1.PFWF);
  BStream:=TBlobStream.Create((Table1.FieldByName('Audio') as TBlobField),bmWrite);
  repeat
  cursize:=readWaveBlock(@buf,sizeof(buf));
  BStream.Write(Buf,curSize);
  until cursize<sizeof(buf);
  CloseWaveFile;
  BStream.Free;
end;

procedure TMAIN.makeRecord(table: TTable);
begin
with Table do begin
  if FindKey([Clndr.CalendarDate]) then Delete;
  Append;
  FieldByName('ACMFormat').asString:=WaveInput1.ACMFormat;
  FieldByName('Date').asString:=DateToStr(Clndr.CalendarDate);
  FieldByName('Format').SetData(WaveInput1.PFWF);
  BStream:=TBlobStream.Create((FieldByName('Audio') as TBlobField),bmWrite);
end;
end;

procedure TMAIN.getAudioFormat(table: TTable);
begin
  with table do begin
    if  WaveOutput1.sizeFWF<>0 then FreeMem(WaveOutput1.PFWF, WaveOutput1.sizeFWF);
    GetMem(WaveOutput1.PFWF,FieldByName('Format').DataSize);
    FieldByName('Format').GetData(WaveOutput1.PFWF);
    WaveOutput1.sizeFWF:=FieldByName('Format').DataSize;
    paFormat.caption:=FieldByName('ACMFormat').asString;
  end;
end;

procedure TMAIN.getAudioData(table: TTable);
begin
   WaveOutput1.Open;
   {set the buffersize to equal wave block size}
   if WaveOutput1.BlockAlign>4 then
        buffersize:=WaveOutput1.PFWF^.nBlockAlign * 5
   else buffersize:=WaveOutput1.BlockAlign*9096;

   {allocate memory for the buffer, create the stream}
   GetMem(buffer,buffersize);
   BStream:=TBlobStream.Create((Table.FieldByName('Audio') as TBlobField),bmRead);

   ID:=1;  {output buffer starting position}
   {load the output buffer from the stream}
   repeat
     cursize:=BStream.Read(buffer^,buffersize);
     WaveOutput1.AddBlock(buffer,cursize,ID);
     Inc(ID);
   until (cursize=0) or (WaveOutput1.MessageCount=WaveOutput1.Capacity);
end;


procedure TMAIN.WaveInput1Mute(Sender: TObject);
begin
 if waveInput1.autoActivation then  {show current mic input level}
   tbActivationLevel.SelHighEnd:=WaveInput1.CurLevel div 10;
   log.lines.add('waveIn mute');
end;

procedure TMAIN.WaveInput1Overflow(Sender: TObject);
begin
 if waveInput1.autoActivation then  {show current mic input level}
   tbActivationLevel.SelHighEnd:=WaveInput1.CurLevel div 10;
   log.lines.add('waveIn buffer overflow');
end;

procedure TMAIN.WaveInput1WaveInput(Sender: TObject);
var WE : TWaveEvent;
begin
 log.lines.add('WaveInput');

 if waveInput1.autoActivation then  {show current mic input level}
   tbActivationLevel.SelHighEnd:=WaveInput1.CurLevel div 10;
   (*take the current wave event from the input buffer,
     and write it to the stream*)

   WE:=WaveInput1.GetWaveEvent;
   if WE=nil then exit;
   if (WE.BufData<>nil) and (WE.BufSize>0) then begin
     BStream.Write(WE.Bufdata^,WE.BufSize);
     WE.Free;
   end
   else WE.Free;
end;

procedure TMAIN.WaveOutput1WaveOutput(Sender: TObject);
begin
   log.lines.add('WaveOutput');
   WaveOutput1.PlayBuffer;   {play the next buffer, if there is one.}
   WaveOutput1.PlayBuffer;
   cursize:=BStream.Read(buffer^,buffersize);  {see if there's more data in the stream}
   if cursize<>0 then  begin                   {if there is, then add it to the output buffer}
      WaveOutput1.AddBlock(buffer,cursize,ID); {so it will be available to be played on the }
      Inc(ID);                                 {next WaveOutput event }
   end;
end;

procedure TMAIN.WaveOutput1End(Sender: TObject);
begin
   log.lines.add('WaveOut-OnEnd');
   paStatus.caption := 'All done!';
   btPlay.Down:=false;
   btPlayClick(Sender);
end;


end.


